import logging
from uuid import UUID
from flask import (
    Blueprint, flash, g, redirect, render_template, request, session, url_for, make_response
)
from argus.backend.controller.notifications import bp as notifications_bp
from argus.backend.service.argus_service import ArgusService
from argus.backend.models.web import WebFileStorage
from argus.backend.service.user import UserService, login_required

LOGGER = logging.getLogger(__name__)

bp = Blueprint('teams', __name__, url_prefix="/teams")

@bp.route("/")
@login_required
def index():
    return render_template("teams.html.j2")



from flask import Blueprint

from argus.backend.plugins.core import PluginInfoBase, PluginModelBase
from argus.backend.plugins.driver_matrix_tests.model import DriverTestRun
from argus.backend.plugins.driver_matrix_tests.udt import TestCollection, EnvironmentInfo, TestCase, TestSuite


class PluginInfo(PluginInfoBase):
    # pylint: disable=too-few-public-methods
    name: str = "driver-matrix-tests"
    model: PluginModelBase = DriverTestRun
    controller: Blueprint = None
    all_models = [
        DriverTestRun,
    ]
    all_types = [
        TestCollection,
        TestSuite,
        TestCase,
        EnvironmentInfo
    ]



from datetime import datetime
import json
import logging
import re
import time
from typing import Any
from uuid import UUID

import requests
from flask import g

from argus.backend.models.web import (
    ArgusEvent,
    ArgusEventTypes,
    ArgusGithubIssue,
    ArgusGroup,
    ArgusNotificationSourceTypes,
    ArgusNotificationTypes,
    ArgusRelease,
    ArgusTest,
    ArgusTestRunComment,
    User,
    UserOauthToken,
)

from argus.backend.plugins.core import PluginInfoBase, PluginModelBase

from argus.backend.plugins.loader import AVAILABLE_PLUGINS
from argus.backend.events.event_processors import EVENT_PROCESSORS
from argus.backend.service.notification_manager import NotificationManagerService
from argus.backend.util.common import get_build_number, strip_html_tags
from argus.backend.util.enums import TestInvestigationStatus, TestStatus

LOGGER = logging.getLogger(__name__)


class TestRunService:
    ASSIGNEE_PLACEHOLDER = "none-none-none"

    RE_MENTION = r"@[A-Za-z\d](?:[A-Za-z\d]|-(?=[A-Za-z\d])){0,38}"

    plugins = AVAILABLE_PLUGINS
    github_headers = {
        "Accept": "application/vnd.github.v3+json",
    }

    def __init__(self) -> None:
        self.notification_manager = NotificationManagerService()

    def get_plugin(self, plugin_name: str) -> PluginInfoBase | None:
        return self.plugins.get(plugin_name)

    def get_run(self, run_type: str, run_id: UUID) -> PluginModelBase:
        plugin = self.plugins.get(run_type)
        if plugin:
            try:
                return plugin.model.get(id=run_id)
            except plugin.model.DoesNotExist:
                return None

    def get_runs_by_test_id(self, test_id: UUID, additional_runs: list[UUID], limit: int = 10):
        test: ArgusTest = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(plugin_name=test.plugin_name)
        if not plugin:
            return []

        last_runs: list[PluginModelBase] = list(plugin.model.filter(build_id=test.build_system_id).limit(limit).all())
        last_runs_ids = [run.id for run in last_runs]
        for added_run in additional_runs:
            if added_run not in last_runs_ids:
                last_runs.append(plugin.model.get(id=added_run))

        for row in last_runs:
            setattr(row, "build_number", get_build_number(build_job_url=row.build_job_url))

        return last_runs

    def get_runs_by_id(self, test_id: UUID, runs: list[UUID]):  # FIXME: Not needed, use get_run and individual polling
        # This is a batch request.
        test = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(plugin_name=test.plugin_name)
        polled_runs: list[PluginModelBase] = []
        for run_id in runs:
            try:
                run: PluginModelBase = plugin.model.get(id=run_id)
                polled_runs.append(run)
            except plugin.model.DoesNotExist:
                pass

        response = {str(run.id): run for run in polled_runs}
        return response

    def change_run_status(self, test_id: UUID, run_id: UUID, new_status: TestStatus):
        test = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(plugin_name=test.plugin_name)
        run: PluginModelBase = plugin.model.get(id=run_id)
        old_status = run.status
        run.status = new_status.value
        run.save()

        self.create_run_event(
            kind=ArgusEventTypes.TestRunStatusChanged,
            body={
                "message": "Status was changed from {old_status} to {new_status} by {username}",
                "old_status": old_status,
                "new_status": new_status.value,
                "username": g.user.username
            },
            user_id=g.user.id,
            run_id=run.id,
            release_id=test.release_id,
            group_id=test.group_id,
            test_id=test.id
        )

        return {
            "test_run_id": run.id,
            "status": new_status
        }

    def change_run_investigation_status(self, test_id: UUID, run_id: UUID, new_status: TestInvestigationStatus):
        test = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(plugin_name=test.plugin_name)
        run: PluginModelBase = plugin.model.get(id=run_id)
        old_status = run.investigation_status
        run.investigation_status = new_status.value
        run.save()

        self.create_run_event(
            kind=ArgusEventTypes.TestRunStatusChanged,
            body={
                "message": "Investigation status was changed from {old_status} to {new_status} by {username}",
                "old_status": old_status,
                "new_status": new_status.value,
                "username": g.user.username
            },
            user_id=g.user.id,
            run_id=run.id,
            release_id=test.release_id,
            group_id=test.group_id,
            test_id=test.id
        )

        return {
            "test_run_id": run.id,
            "investigation_status": new_status
        }

    def change_run_assignee(self, test_id: UUID, run_id: UUID, new_assignee: UUID | None):
        test = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(plugin_name=test.plugin_name)
        run: PluginModelBase = plugin.model.get(id=run_id)
        old_assignee = run.assignee
        run.assignee = new_assignee
        run.save()

        if new_assignee:
            new_assignee_user = User.get(id=new_assignee)
        else:
            new_assignee_user = None
        if old_assignee:
            try:
                old_assignee_user = User.get(id=old_assignee)
            except User.DoesNotExist:
                LOGGER.warning("Non existent assignee was present on the run %s for test %s: %s",
                               run_id, test_id, old_assignee)
                old_assignee = None
        self.create_run_event(
            kind=ArgusEventTypes.AssigneeChanged,
            body={
                "message": "Assignee was changed from \"{old_user}\" to \"{new_user}\" by {username}",
                "old_user": old_assignee_user.username if old_assignee else "None",
                "new_user": new_assignee_user.username if new_assignee else "None",
                "username": g.user.username
            },
            user_id=g.user.id,
            run_id=run.id,
            release_id=test.release_id,
            group_id=test.group_id,
            test_id=test.id
        )
        return {
            "test_run_id": run.id,
            "assignee": str(new_assignee_user.id) if new_assignee_user else None
        }

    def get_run_comment(self, comment_id: UUID):
        try:
            return ArgusTestRunComment.get(id=comment_id)
        except ArgusTestRunComment.DoesNotExist:
            return None

    def get_run_comments(self, run_id: UUID):
        return sorted(ArgusTestRunComment.filter(test_run_id=run_id).all(), key=lambda c: c.posted_at)

    def post_run_comment(self, test_id: UUID, run_id: UUID, message: str, reactions: dict, mentions: list[str]):
        message_stripped = strip_html_tags(message)

        mentions = set(mentions)
        for potential_mention in re.findall(self.RE_MENTION, message_stripped):
            if user := User.exists_by_name(potential_mention.lstrip("@")):
                mentions.add(user)

        test: ArgusTest = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(test.plugin_name)
        release: ArgusRelease = ArgusRelease.get(id=test.release_id)
        comment = ArgusTestRunComment()
        comment.message = message_stripped
        comment.reactions = reactions
        comment.mentions = [m.id for m in mentions]
        comment.test_run_id = run_id
        comment.release_id = release.id
        comment.user_id = g.user.id
        comment.posted_at = time.time()
        comment.save()

        run: PluginModelBase = plugin.model.get(id=run_id)
        build_number = get_build_number(build_job_url=run.build_job_url)
        for mention in mentions:
            params = {
                "username": g.user.username,
                "run_id": comment.test_run_id,
                "test_id": test.id,
                "build_id": run.build_id,
                "build_number": build_number,
            }
            self.notification_manager.send_notification(
                receiver=mention.id,
                sender=comment.user_id,
                notification_type=ArgusNotificationTypes.Mention,
                source_type=ArgusNotificationSourceTypes.Comment,
                source_id=comment.id,
                source_message=comment.message,
                content_params=params
            )

        self.create_run_event(kind=ArgusEventTypes.TestRunCommentPosted, body={
            "message": "A comment was posted by {username}",
            "username": g.user.username
        }, user_id=g.user.id, run_id=run_id, release_id=release.id, test_id=test.id)

        return self.get_run_comments(run_id=run_id)

    def delete_run_comment(self, comment_id: UUID, test_id: UUID, run_id: UUID):
        comment: ArgusTestRunComment = ArgusTestRunComment.get(id=comment_id)
        if comment.user_id != g.user.id:
            raise Exception("Unable to delete other user comments")
        comment.delete()

        self.create_run_event(kind=ArgusEventTypes.TestRunCommentDeleted, body={
            "message": "A comment was deleted by {username}",
            "username": g.user.username
        }, user_id=g.user.id, run_id=run_id, release_id=comment.release_id, test_id=test_id)

        return self.get_run_comments(run_id=run_id)

    def update_run_comment(self, comment_id: UUID, test_id: UUID, run_id: UUID, message: str, mentions: list[str], reactions: dict):
        comment: ArgusTestRunComment = ArgusTestRunComment.get(id=comment_id)
        if comment.user_id != g.user.id:
            raise Exception("Unable to edit other user comments")
        comment.message = strip_html_tags(message)
        comment.reactions = reactions
        comment.mentions = mentions
        comment.save()

        self.create_run_event(kind=ArgusEventTypes.TestRunCommentUpdated, body={
            "message": "A comment was edited by {username}",
            "username": g.user.username
        }, user_id=g.user.id, run_id=run_id, release_id=comment.release_id, test_id=test_id)

        return self.get_run_comments(run_id=run_id)

    def create_run_event(self, kind: ArgusEventTypes, body: dict, user_id=None, run_id=None, release_id=None, group_id=None, test_id=None):
        event = ArgusEvent()
        event.release_id = release_id
        event.group_id = group_id
        event.test_id = test_id
        event.user_id = user_id
        event.run_id = run_id
        event.body = json.dumps(body, ensure_ascii=True, separators=(',', ':'))
        event.kind = kind.value
        event.created_at = datetime.utcnow()
        event.save()

    def get_run_events(self, run_id: UUID):
        response = {}
        all_events = ArgusEvent.filter(run_id=run_id).all()
        all_events = sorted(all_events, key=lambda ev: ev.created_at)
        response["run_id"] = run_id
        response["raw_events"] = [dict(event.items()) for event in all_events]
        response["events"] = {
            str(event.id): EVENT_PROCESSORS.get(event.kind)(json.loads(event.body))
            for event in all_events
        }
        return response

    def submit_github_issue(self, issue_url: str, test_id: UUID, run_id: UUID):
        user_tokens = UserOauthToken.filter(user_id=g.user.id).all()
        token = None
        for tok in user_tokens:
            if tok.kind == "github":
                token = tok.token
                break
        if not token:
            raise Exception("Github token not found")

        match = re.match(
            r"http(s)?://(www\.)?github\.com/(?P<owner>[\w\d]+)/"
            r"(?P<repo>[\w\d\-_]+)/(?P<type>issues|pull)/(?P<issue_number>\d+)(/)?",
            issue_url,
        )
        if not match:
            raise Exception("URL doesn't match Github schema")

        test: ArgusTest = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(plugin_name=test.plugin_name)

        run = plugin.model.get(id=run_id)
        release = ArgusRelease.get(id=run["release_id"])
        test = ArgusTest.get(build_system_id=run["build_id"])
        group = ArgusGroup.get(id=test.group_id)

        new_issue = ArgusGithubIssue()
        new_issue.user_id = g.user.id
        new_issue.run_id = run_id
        new_issue.group_id = group.id
        new_issue.release_id = release.id
        new_issue.test_id = test.id
        new_issue.type = match.group("type")
        new_issue.owner = match.group("owner")
        new_issue.repo = match.group("repo")
        new_issue.issue_number = int(match.group("issue_number"))

        issue_request = requests.get(
            f"https://api.github.com/repos/{new_issue.owner}/{new_issue.repo}/issues/{new_issue.issue_number}",
            headers={
                **self.github_headers,
                "Authorization": f"token {token}",
            }
        )
        if issue_request.status_code != 200:
            raise Exception(f"Error getting issue state: Response: HTTP {issue_request.status_code}", issue_request)

        issue_state: dict[str, Any] = issue_request.json()

        new_issue.title = issue_state.get("title")
        new_issue.url = issue_state.get("html_url")
        new_issue.last_status = issue_state.get("state")
        new_issue.save()

        self.create_run_event(
            kind=ArgusEventTypes.TestRunIssueAdded,
            body={
                "message": "An issue titled \"{title}\" was added by {username}",
                "username": g.user.username,
                "url": issue_url,
                "title": issue_state.get("title"),
                "state": issue_state.get("state"),
            },
            user_id=g.user.id,
            run_id=new_issue.run_id,
            release_id=new_issue.release_id,
            group_id=new_issue.group_id,
            test_id=new_issue.test_id
        )

        response = {
            **dict(list(new_issue.items())),
            "title": issue_state.get("title"),
            "state": issue_state.get("state"),
        }

        return response

    def get_github_issues(self, filter_key: str, filter_id: UUID, aggregate_by_issue: bool = False) -> dict:
        if filter_key not in ["release_id", "group_id", "test_id", "run_id", "user_id"]:
            raise Exception(
                "filter_key can only be one of: \"release_id\", \"group_id\", \"test_id\", \"run_id\", \"user_id\""
            )

        all_issues = ArgusGithubIssue.filter(**{filter_key: filter_id}).all()
        if aggregate_by_issue:
            runs_by_issue = {}
            response = []
            for issue in all_issues:
                runs = runs_by_issue.get(issue, [])
                runs.append(issue.run_id)
                runs_by_issue[issue] = runs

            for issue, runs in runs_by_issue.items():
                issue_dict = dict(issue.items())
                issue_dict["runs"] = runs
                response.append(issue_dict)

        else:
            response = [dict(issue.items()) for issue in all_issues]
        return response

    def delete_github_issue(self, issue_id: UUID) -> dict:
        issue: ArgusGithubIssue = ArgusGithubIssue.get(id=issue_id)

        self.create_run_event(
            kind=ArgusEventTypes.TestRunIssueRemoved,
            body={
                "message": "An issue titled \"{title}\" was removed by {username}",
                "username": g.user.username,
                "url": issue.url,
                "title": issue.title,
                "state": issue.last_status,
            },
            user_id=g.user.id,
            run_id=issue.run_id,
            release_id=issue.release_id,
            group_id=issue.group_id,
            test_id=issue.test_id
        )
        issue.delete()

        return {
            "deleted": issue_id
        }



import logging
from typing import Any
from uuid import UUID
from flask import current_app, render_template
from argus.backend.models.web import ArgusNotification, ArgusNotificationSourceTypes, ArgusNotificationTypes, ArgusNotificationState, User
from argus.backend.util.send_email import Email

LOGGER = logging.getLogger(__name__)


class NotificationManagerException(Exception):
    pass


class NotificationManagerService:

    def __init__(self, notification_senders: list["NotificationSenderBase"] | None = None) -> None:
        self.notification_services: list["NotificationSenderBase"] = [
            ArgusDBNotificationSaver(),
            EmailNotificationServiceSender()
        ]
        if notification_senders:
            self.notification_services.extend(notification_senders)

    def send_notification(self, receiver: UUID, sender: UUID,
                          notification_type: ArgusNotificationTypes,
                          source_type: ArgusNotificationSourceTypes,
                          source_id: UUID, title: str | None = None,
                          source_message: str | None = None, content_params: dict | None = None) -> Any:

        for service in self.notification_services:
            service.send_notification(receiver, sender, notification_type, source_type,
                                      source_id, title, source_message, content_params)

    def get_notificaton(self, receiver: UUID, notification_id: UUID) -> ArgusNotification:
        return ArgusNotification.get(receiver=receiver, id=notification_id)

    def get_unread_count(self, receiver: UUID) -> int:
        query = ArgusNotification.filter(
            receiver=receiver, state__eq=ArgusNotificationState.UNREAD.value).allow_filtering().all()
        return len(query)

    def read_notification(self, receiver: UUID, notification_id: UUID) -> ArgusNotification:
        notification = ArgusNotification.get(receiver=receiver, id=notification_id)
        notification.state = ArgusNotificationState.READ
        return bool(notification.save().state)

    def get_notifications(self, receiver: UUID, limit: int = 20, after: UUID | None = None) -> list[ArgusNotification]:
        if after:
            return ArgusNotification.filter(receiver=receiver, id__lte=after).all().limit(limit)
        return ArgusNotification.filter(receiver=receiver).all().limit(limit)


class NotificationSenderBase:
    NOTIFICATION_TITLES = {
        ArgusNotificationTypes.Mention: "You were mentioned in a comment",
        ArgusNotificationTypes.StatusChange: "A run you are assigned to changed status",
        ArgusNotificationTypes.AssigneeChange: "You were assigned to a run",
        ArgusNotificationTypes.ScheduleChange: "You were assigned to a schedule",
    }

    CONTENT_TEMPLATES = {}

    def _check_user(self, user_id: UUID) -> bool:
        try:
            user = User.get(id=user_id)
            if user.id == user_id:
                return True
        except User.DoesNotExist:
            return False

    def _get_title_for_notification_type(self, notification_type: ArgusNotificationTypes) -> str:
        if title := self.NOTIFICATION_TITLES.get(notification_type):
            return title
        raise NotificationManagerException(
            f"Title for notification type {notification_type} not found.", notification_type)

    def _render_content(self, content_type: ArgusNotificationTypes, params: dict):
        if content_renderer := self.CONTENT_TEMPLATES.get(content_type):
            return content_renderer(params)
        raise NotificationManagerException(
            f"Content renderer for notification type {content_type} not found.", content_type)

    def send_notification(self, receiver: UUID, sender: UUID,
                          notification_type: ArgusNotificationTypes,
                          source_type: ArgusNotificationSourceTypes,
                          source_id: UUID,
                          title: str | None = None,
                          content: str | None = None,
                          content_params: dict | None = None) -> Any:
        raise NotImplementedError()


class ArgusDBNotificationSaver(NotificationSenderBase):
    CONTENT_TEMPLATES = {
        ArgusNotificationTypes.Mention: lambda p: render_template("notifications/mention.html.j2", **p if p else {})
    }

    def send_notification(self, receiver: UUID, sender: UUID, notification_type: ArgusNotificationTypes, source_type: ArgusNotificationSourceTypes,
                          source_id: UUID, title: str | None = None, content: str | None = None, content_params: dict | None = None) -> ArgusNotification:
        new_notification = ArgusNotification()
        for user in [sender, receiver]:
            if not self._check_user(user_id=user):
                raise NotificationManagerException(f"UserId {user} not found in the database", user)

        new_notification.sender = sender
        new_notification.receiver = receiver
        new_notification.type = notification_type.value
        new_notification.source_type = source_type.value
        new_notification.source_id = source_id
        new_notification.title = title if title else self._get_title_for_notification_type(notification_type)
        new_notification.content = content if not content_params else self._render_content(
            notification_type, content_params)
        return new_notification.save()


class EmailNotificationServiceSender(NotificationSenderBase):
    CONTENT_TEMPLATES = {
        ArgusNotificationTypes.Mention: lambda p: render_template(
            "notifications/email_mention.html.j2", **p if p else {})
    }

    def __init__(self):
        self.email = Email(init_connection=False)

    def get_user(self, user_id: UUID) -> User:
        return User.get(id=user_id)

    def send_notification(self, receiver: UUID, sender: UUID,
                          notification_type: ArgusNotificationTypes, source_type: ArgusNotificationSourceTypes,
                          source_id: UUID, title: str | None = None, content: str | None = None, content_params: dict | None = None):
        try:
            content_params = content_params or {}
            receiver_user = self.get_user(receiver)
            sender_user = self.get_user(sender)
            subject = title if title else self._get_title_for_notification_type(notification_type)
            content_params.update({
                                  "sender": sender_user.full_name,
                                  "message": content,
                                  })

            email_content = self._render_content(notification_type, content_params)

            self.email.send(subject=f"Argus Notification - {subject}",
                            content=email_content,
                            recipients=[receiver_user.email])
        except Exception as details:  # pylint: disable=broad-except
            current_app.logger.error("Failed to send email: %s", details)




from flask import g, current_app, session

from argus.backend.db import ScyllaCluster
from argus.backend.models.web import (
    ArgusGithubIssue,
    ArgusRelease,
    ArgusGroup,
    ArgusTest,
    ArgusSchedule,
    ArgusScheduleAssignee,
    ArgusScheduleGroup,
    ArgusScheduleTest,
    ArgusTestRunComment,
    ArgusEvent,
    ArgusEventTypes,
    ReleasePlannerComment,
    User,
    UserOauthToken,
    WebFileStorage,
)


class AdminService:
    def __init__(self, database_session=None):
        self.session = database_session if database_session else ScyllaCluster.get_session()
        self.database = ScyllaCluster.get()



import logging
from traceback import format_exception
from flask import request

LOGGER = logging.getLogger(__name__)


class APIException(Exception):
    pass


def handle_api_exception(exception: Exception):
    LOGGER.error("Exception in %s\n%s", request.endpoint, "".join(format_exception(exception)))

    return {
        "status": "error",
        "response": {
            "exception": exception.__class__.__name__,
            "arguments": exception.args
        }
    }



from functools import partial
from datetime import datetime

from argus.backend.models.web import User
from argus.backend.util.module_loaders import is_filter, export_functions


export_filters = partial(export_functions, module_name=__name__, attr="is_filter")


@is_filter("from_timestamp")
def from_timestamp_filter(timestamp: int):
    return datetime.utcfromtimestamp(timestamp)


@is_filter("safe_user")
def safe_user(user: User):
    user_dict = dict(user.items())
    del user_dict["password"]
    return user_dict



import logging
from typing import Callable
from uuid import UUID

from flask import Request, Response


LOGGER = logging.getLogger(__name__)
FlaskView = Callable[..., Response]


def first(iterable, value, key: Callable = None, predicate: Callable = None):
    for elem in iterable:
        if predicate and predicate(elem, value):
            return elem
        elif key and key(elem) == value:
            return elem
        elif elem == value:
            return elem
    return None


def check_scheduled_test(test, group, testname):
    return testname in (f"{group}/{test}", test)


def strip_html_tags(text: str):
    return text.replace("<", "&lt;").replace(">", "&gt;")


def convert_str_list_to_uuid(lst: list[str]) -> list[UUID]:
    return [UUID(s) for s in lst]


def get_payload(client_request: Request) -> dict:
    if not client_request.is_json:
        raise Exception(
            "Content-Type mismatch, expected application/json, got:",
            client_request.content_type
        )
    request_payload = client_request.get_json()

    return request_payload


def get_build_number(build_job_url: str) -> int | None:
    build_number = build_job_url.rstrip("/").split("/")[-1] if build_job_url else -1
    if build_number:
        try:
            return int(build_number)
        except ValueError:
            LOGGER.error("Error parsing build number from %s: got %s as build_number", build_job_url, build_number)
    return None



from dataclasses import dataclass
from datetime import datetime
from uuid import UUID
from cassandra.cqlengine import columns
from argus.backend.db import ScyllaCluster
from argus.backend.models.web import ArgusRelease
from argus.backend.plugins.core import PluginModelBase
from argus.backend.plugins.driver_matrix_tests.udt import TestCollection, TestSuite, TestCase, EnvironmentInfo
from argus.backend.plugins.driver_matrix_tests.raw_types import RawMatrixTestResult


@dataclass(init=True, repr=True, frozen=True)
class DriverMatrixRunSubmissionRequest():
    schema_version: str
    run_id: str
    job_name: str
    job_url: str
    test_environment: dict[str, str]
    matrix_results: list[RawMatrixTestResult]


class DriverTestRun(PluginModelBase):
    _plugin_name = "driver-matrix-tests"
    __table_name__ = "driver_test_run"
    scylla_version = columns.Text()
    test_collection = columns.List(value_type=columns.UserDefinedType(user_type=TestCollection))
    environment_info = columns.List(value_type=columns.UserDefinedType(user_type=EnvironmentInfo))

    @classmethod
    def _stats_query(cls) -> str:
        return ("SELECT id, test_id, group_id, release_id, status, start_time, build_job_url, build_id, "
                f"assignee, end_time, investigation_status, heartbeat, scylla_version FROM {cls.table_name()} WHERE release_id = ?")

    @classmethod
    def get_distinct_product_versions(cls, release: ArgusRelease) -> list[str]:
        cluster = ScyllaCluster.get()
        statement = cluster.prepare(f"SELECT scylla_version FROM {cls.table_name()} WHERE release_id = ?")
        rows = cluster.session.execute(query=statement, parameters=(release.id,))
        unique_versions = {r["scylla_version"] for r in rows if r["scylla_version"]}

        return sorted(list(unique_versions), reverse=True)

    @classmethod
    def load_test_run(cls, run_id: UUID) -> 'DriverTestRun':
        return cls.get(id=run_id)

    @classmethod
    def submit_run(cls, request_data: dict) -> 'DriverTestRun':
        req = DriverMatrixRunSubmissionRequest(**request_data)
        run = cls()
        run.id = req.run_id  # pylint: disable=invalid-name
        run.build_id = req.job_name
        run.build_job_url = req.job_url
        run.assign_categories()
        for key, value in req.test_environment.items():
            env_info = EnvironmentInfo()
            env_info.key = key
            env_info.value = value
            run.environment_info.append(env_info)

        run.scylla_version = req.test_environment.get("scylla-version").replace("~", ".")
        run.test_collection = []

        for result in req.matrix_results:
            collection = TestCollection()
            collection.name = result.get("name")
            collection.driver = result.get("driver_name")
            collection.tests_total = result.get("tests")
            collection.failures = result.get("failures")
            collection.errors = result.get("errors")
            collection.skipped = result.get("skipped")
            collection.passed = result.get("passed")
            collection.disabled = result.get("disabled")
            collection.time = result.get("time")
            collection.timestamp = datetime.fromisoformat(result.get("timestamp"))
            for raw_suite in result.get("suites"):
                suite = TestSuite()
                suite.name = raw_suite.get("name")
                suite.tests_total = raw_suite.get("tests")
                suite.failures = raw_suite.get("failures")
                suite.errors = raw_suite.get("errors")
                suite.skipped = raw_suite.get("skipped")
                suite.passed = raw_suite.get("passed")
                suite.disabled = raw_suite.get("disabled")
                suite.time = raw_suite.get("time")

                for raw_case in raw_suite.get("cases"):
                    case = TestCase()
                    case.name = raw_case.get("name")
                    case.status = raw_case.get("status")
                    case.time = raw_case.get("time")
                    case.classname = raw_case.get("classname")
                    case.message = raw_case.get("message")
                    suite.cases.append(case)

                collection.suites.append(suite)
            run.test_collection.append(collection)

        run.save()
        return run

    def get_resources(self) -> list:
        return []

    def get_nemeses(self) -> list:
        return []

    def get_events(self) -> list:
        return []

    def submit_product_version(self, version: str):
        self.scylla_version = version

    def finish_run(self):
        self.end_time = datetime.utcnow()

    def submit_logs(self, logs: list[dict]):
        pass



import logging
import click
from flask import Blueprint
from flask.cli import with_appcontext
from cassandra.cqlengine.management import sync_table, sync_type
from argus.backend.db import ScyllaCluster
from argus.backend.plugins.loader import all_plugin_models, all_plugin_types
from argus.backend.service.build_system_monitor import JenkinsMonitor

cli_bp = Blueprint("cli", __name__)
LOGGER = logging.getLogger(__name__)


@cli_bp.cli.add_command
@click.command('sync-models')
@with_appcontext
def sync_models_command():
    cluster = ScyllaCluster.get()
    cluster.sync_core_tables()
    LOGGER.info("Synchronizing plugin types...")
    for user_type in all_plugin_types():
        LOGGER.info("Synchronizing plugin type %s...", user_type.__name__)
        sync_type(ks_name=cluster.config["SCYLLA_KEYSPACE_NAME"], type_model=user_type)
    LOGGER.info("Synchronizing plugin models...")
    for model in all_plugin_models():
        LOGGER.info("Synchronizing plugin model %s...", model.__name__)
        sync_table(model=model, keyspaces=[cluster.config["SCYLLA_KEYSPACE_NAME"]])

    LOGGER.info("Plugins ready.")
    click.echo("All models synchronized.")


@cli_bp.cli.add_command
@click.command('scan-jenkins')
@with_appcontext
def scan_jenkins_command():
    monitor = JenkinsMonitor()
    monitor.collect()
    click.echo("Done.")



import ipaddress
import time
from enum import Enum
from typing import Optional
from pydantic.dataclasses import dataclass
from pydantic import ValidationError, validator
from argus.db.db_types import ArgusUDTBase


@dataclass(init=True, repr=True)
class CloudInstanceDetails(ArgusUDTBase):
    provider: str = ""
    region: str = ""
    public_ip: str = ""
    private_ip: str = ""
    creation_time: int = 0
    termination_time: int = 0
    termination_reason: str = ""
    shards_amount: Optional[int] = 0
    _typename = "CloudInstanceDetails_v3"

    @classmethod
    def from_db_udt(cls, udt):
        return cls(provider=udt.provider, region=udt.region, public_ip=udt.public_ip, private_ip=udt.private_ip,
                   creation_time=udt.creation_time, termination_time=udt.termination_time,
                   termination_reason=udt.termination_reason, shards_amount=udt.shards_amount)

    @validator("public_ip")
    def valid_public_ip_address(cls, v):
        try:
            ipaddress.ip_address(v)
        except ValueError:
            raise ValidationError(f"Not a valid IPv4(v6) address: {v}")

        return v

    @validator("private_ip")
    def valid_private_ip_address(cls, v):
        try:
            ipaddress.ip_address(v)
        except ValueError:
            raise ValidationError(f"Not a valid IPv4(v6) address: {v}")

        return v


@dataclass(init=True, repr=True)
class CloudNodesInfo(ArgusUDTBase):
    image_id: str
    instance_type: str
    node_amount: int
    post_behaviour: str

    @classmethod
    def from_db_udt(cls, udt):
        return cls(image_id=udt.image_id, instance_type=udt.instance_type,
                   node_amount=udt.node_amount, post_behaviour=udt.post_behaviour)


@dataclass(init=True, repr=True)
class BaseCloudSetupDetails(ArgusUDTBase):
    db_node: CloudNodesInfo
    loader_node: CloudNodesInfo
    monitor_node: CloudNodesInfo
    backend: str = None
    _typename = "CloudSetupDetails"

    @classmethod
    def from_db_udt(cls, udt):
        db_node = CloudNodesInfo(*udt.db_node)
        loader_node = CloudNodesInfo(*udt.loader_node)
        monitor_node = CloudNodesInfo(*udt.monitor_node)
        backend = udt.backend
        return cls(db_node=db_node, loader_node=loader_node, monitor_node=monitor_node, backend=backend)


@dataclass(init=True, repr=True)
class AWSSetupDetails(BaseCloudSetupDetails):
    backend: str = "aws"


@dataclass(init=True, repr=True)
class GCESetupDetails(BaseCloudSetupDetails):
    backend: str = "gce"


class ResourceState(str, Enum):
    RUNNING = "running"
    STOPPED = "stopped"
    TERMINATED = "terminated"


@dataclass(init=True, repr=True)
class CloudResource(ArgusUDTBase):
    name: str
    state: ResourceState
    resource_type: str
    instance_info: CloudInstanceDetails
    _typename = "CloudResource_v3"

    def __eq__(self, other) -> bool:
        if (isinstance(other, CloudResource)):
            return self.name == other.name
        return False

    @classmethod
    def from_db_udt(cls, udt):
        instance_info = CloudInstanceDetails.from_db_udt(udt.instance_info)
        return cls(name=udt.name, state=udt.state, resource_type=udt.resource_type, instance_info=instance_info)

    @validator("state")
    def valid_state(cls, v):
        try:
            ResourceState(v)
        except ValueError:
            raise ValidationError(f"Not a valid resource state: {v}")
        return v

    def terminate(self, reason):
        self.state = ResourceState.TERMINATED
        self.instance_info.termination_time = int(time.time())
        self.instance_info.termination_reason = reason

    def stop(self):
        self.state = ResourceState.STOPPED



from argus.backend.models.web import ArgusEventTypes


def event_process_posted_comment(event: dict) -> dict:
    return event["message"].format(**event)


def event_process_status_changed(event: dict) -> dict:
    return event["message"].format(**event)


def event_process_investigation_status_changed(event: dict) -> dict:
    return event["message"].format(**event)


def event_process_assignee_changed(event: dict) -> dict:
    return event["message"].format(**event)


def event_process_issue_added(event: dict) -> dict:
    return event["message"].format(**event)


EVENT_PROCESSORS = {
    ArgusEventTypes.AssigneeChanged: event_process_assignee_changed,
    ArgusEventTypes.TestRunStatusChanged: event_process_status_changed,
    ArgusEventTypes.TestRunCommentPosted: event_process_posted_comment,
    ArgusEventTypes.TestRunCommentUpdated: event_process_posted_comment,
    ArgusEventTypes.TestRunCommentDeleted: event_process_posted_comment,
    ArgusEventTypes.TestRunIssueAdded: event_process_issue_added,
    ArgusEventTypes.TestRunIssueRemoved: event_process_issue_added,
    ArgusEventTypes.TestRunInvestigationStatusChanged: event_process_investigation_status_changed,
}



import logging
from argus.backend.plugins.sct.udt import CloudNodesInfo, CloudSetupDetails

LOGGER = logging.getLogger(__name__)


def _get_node_amounts(config: dict) -> tuple[int, int]:
    num_db_node = config.get("n_db_nodes")
    num_db_node = sum([int(i) for i in num_db_node.split()]) if isinstance(num_db_node, str) else num_db_node
    num_loaders = config.get("n_loaders")
    num_loaders = sum([int(i) for i in num_loaders.split()]) if isinstance(num_loaders, str) else num_loaders

    return num_db_node, num_loaders


def _prepare_aws_resource_setup(sct_config: dict) -> CloudSetupDetails:
    num_db_nodes, n_loaders = _get_node_amounts(sct_config)
    db_node_setup = CloudNodesInfo(image_id=sct_config.get("ami_id_db_scylla"),
                                   instance_type=sct_config.get("instance_type_db"),
                                   node_amount=num_db_nodes,
                                   post_behaviour=sct_config.get("post_behavior_db_nodes"))
    loader_node_setup = CloudNodesInfo(image_id=sct_config.get("ami_id_loader"),
                                       instance_type=sct_config.get("instance_type_loader"),
                                       node_amount=n_loaders,
                                       post_behaviour=sct_config.get("post_behavior_loader_nodes"))
    monitor_node_setup = CloudNodesInfo(image_id=sct_config.get("ami_id_monitor"),
                                        instance_type=sct_config.get("instance_type_monitor"),
                                        node_amount=sct_config.get("n_monitor_nodes"),
                                        post_behaviour=sct_config.get("post_behavior_monitor_nodes"))
    cloud_setup = CloudSetupDetails(db_node=db_node_setup, loader_node=loader_node_setup,
                                    monitor_node=monitor_node_setup, backend=sct_config.get("cluster_backend"))

    return cloud_setup


def _prepare_gce_resource_setup(sct_config: dict) -> CloudSetupDetails:
    num_db_nodes, n_loaders = _get_node_amounts(sct_config)
    db_node_setup = CloudNodesInfo(image_id=sct_config.get("gce_image_db"),
                                   instance_type=sct_config.get("gce_instance_type_db"),
                                   node_amount=num_db_nodes,
                                   post_behaviour=sct_config.get("post_behavior_db_nodes"))
    loader_node_setup = CloudNodesInfo(image_id=sct_config.get("gce_image_loader"),
                                       instance_type=sct_config.get("gce_instance_type_loader"),
                                       node_amount=n_loaders,
                                       post_behaviour=sct_config.get("post_behavior_loader_nodes"))
    monitor_node_setup = CloudNodesInfo(image_id=sct_config.get("gce_image_monitor"),
                                        instance_type=sct_config.get("gce_instance_type_monitor"),
                                        node_amount=sct_config.get("n_monitor_nodes"),
                                        post_behaviour=sct_config.get("post_behavior_monitor_nodes"))
    cloud_setup = CloudSetupDetails(db_node=db_node_setup, loader_node=loader_node_setup,
                                    monitor_node=monitor_node_setup, backend=sct_config.get("cluster_backend"))

    return cloud_setup


def _prepare_azure_resource_setup(sct_config: dict) -> CloudSetupDetails:
    num_db_nodes, n_loaders = _get_node_amounts(sct_config)
    db_node_setup = CloudNodesInfo(image_id=sct_config.get("azure_image_db"),
                                   instance_type=sct_config.get("azure_instance_type_db"),
                                   node_amount=num_db_nodes,
                                   post_behaviour=sct_config.get("post_behavior_db_nodes"))
    loader_node_setup = CloudNodesInfo(image_id=sct_config.get("azure_image_loader"),
                                       instance_type=sct_config.get("azure_instance_type_loader"),
                                       node_amount=n_loaders,
                                       post_behaviour=sct_config.get("post_behavior_loader_nodes"))
    monitor_node_setup = CloudNodesInfo(image_id=sct_config.get("azure_image_monitor"),
                                        instance_type=sct_config.get("azure_instance_type_monitor"),
                                        node_amount=sct_config.get("n_monitor_nodes"),
                                        post_behaviour=sct_config.get("post_behavior_monitor_nodes"))
    cloud_setup = CloudSetupDetails(db_node=db_node_setup, loader_node=loader_node_setup,
                                    monitor_node=monitor_node_setup, backend=sct_config.get("cluster_backend"))

    return cloud_setup


def _prepare_unknown_resource_setup(sct_config: dict) -> CloudSetupDetails:
    LOGGER.error("Unknown backend encountered: %s", sct_config.get("cluster_backend"))
    db_node_setup = CloudNodesInfo(image_id="UNKNOWN",
                                   instance_type="UNKNOWN",
                                   node_amount=-1,
                                   post_behaviour="UNKNOWN")
    loader_node_setup = CloudNodesInfo(image_id="UNKNOWN",
                                       instance_type="UNKNOWN",
                                       node_amount=-1,
                                       post_behaviour="UNKNOWN")
    monitor_node_setup = CloudNodesInfo(image_id="UNKNOWN",
                                        instance_type="UNKNOWN",
                                        node_amount=-1,
                                        post_behaviour="UNKNOWN")
    cloud_setup = CloudSetupDetails(db_node=db_node_setup, loader_node=loader_node_setup,
                                    monitor_node=monitor_node_setup, backend=sct_config.get("cluster_backend"))

    return cloud_setup


def _prepare_bare_metal_resource_setup(sct_config: dict) -> CloudSetupDetails:
    db_node_setup = CloudNodesInfo(image_id="bare_metal",
                                   instance_type="bare_metal",
                                   node_amount=sct_config.get("n_db_nodes"),
                                   post_behaviour=sct_config.get("post_behavior_db_nodes"))
    loader_node_setup = CloudNodesInfo(image_id="bare_metal",
                                       instance_type="bare_metal",
                                       node_amount=sct_config.get("n_loaders"),
                                       post_behaviour=sct_config.get("post_behavior_loader_nodes"))
    monitor_node_setup = CloudNodesInfo(image_id="bare_metal",
                                        instance_type="bare_metal",
                                        node_amount=sct_config.get("n_monitor_nodes"),
                                        post_behaviour=sct_config.get("post_behavior_monitor_nodes"))
    cloud_setup = CloudSetupDetails(db_node=db_node_setup, loader_node=loader_node_setup,
                                    monitor_node=monitor_node_setup, backend=sct_config.get("cluster_backend"))

    return cloud_setup


def _prepare_k8s_gce_minikube_resource_setup(sct_config: dict) -> CloudSetupDetails:
    cloud_setup = _prepare_gce_resource_setup(sct_config)

    image_id = sct_config.get("scylla_version")
    cloud_setup.db_node.image_id = f"scylladb/scylladb:{image_id}"
    cloud_setup.db_node.instance_type = sct_config.get("gce_instance_type_minikube")

    return cloud_setup


def _prepare_k8s_gke_resource_setup(sct_config: dict) -> CloudSetupDetails:
    cloud_setup = _prepare_gce_resource_setup(sct_config)
    image_id = sct_config.get("scylla_version")
    cloud_setup.db_node.image_id = f"scylladb/scylladb:{image_id}"
    cloud_setup.monitor_node.image_id = sct_config.get("mgmt_docker_image")
    cloud_setup.loader_node.image_id = f"scylladb/scylladb:{image_id}"

    return cloud_setup


def _prepare_k8s_eks_resource_setup(sct_config: dict) -> CloudSetupDetails:
    cloud_setup = _prepare_aws_resource_setup(sct_config)

    return cloud_setup


def _prepare_docker_resource_setup(sct_config: dict) -> CloudSetupDetails:
    db_node_setup = CloudNodesInfo(image_id=sct_config.get('docker_image'),
                                   instance_type="docker",
                                   node_amount=sct_config.get("n_db_nodes"),
                                   post_behaviour=sct_config.get("post_behavior_db_nodes"))
    loader_node_setup = CloudNodesInfo(image_id=sct_config.get('docker_image'),
                                       instance_type="docker",
                                       node_amount=sct_config.get("n_loaders"),
                                       post_behaviour=sct_config.get("post_behavior_loader_nodes"))
    monitor_node_setup = CloudNodesInfo(image_id=sct_config.get('docker_image'),
                                        instance_type="docker",
                                        node_amount=sct_config.get("n_monitor_nodes"),
                                        post_behaviour=sct_config.get("post_behavior_monitor_nodes"))
    cloud_setup = CloudSetupDetails(db_node=db_node_setup, loader_node=loader_node_setup,
                                    monitor_node=monitor_node_setup, backend=sct_config.get("cluster_backend"))

    return cloud_setup


class ResourceSetup:
    # pylint: disable=too-few-public-methods
    BACKEND_MAP = {
        "aws": _prepare_aws_resource_setup,
        "aws-siren": _prepare_aws_resource_setup,
        "azure": _prepare_azure_resource_setup,
        "gce": _prepare_gce_resource_setup,
        "gce-siren": _prepare_gce_resource_setup,
        "k8s-eks": _prepare_k8s_eks_resource_setup,
        "k8s-gke": _prepare_k8s_gke_resource_setup,
        "k8s-gce-minikube": _prepare_k8s_gce_minikube_resource_setup,
        "baremetal": _prepare_bare_metal_resource_setup,
        "docker": _prepare_docker_resource_setup,
        "unknown": _prepare_unknown_resource_setup,
    }

    @classmethod
    def get_resource_setup(cls, backend: str, sct_config: dict) -> CloudSetupDetails:
        return cls.BACKEND_MAP.get(backend, _prepare_unknown_resource_setup)(sct_config)



import logging
from logging.config import dictConfig
from flask import has_request_context, request

# pylint: disable=line-too-long
LOG_FORMAT_REQUEST = "[%(levelcolor)s%(levelname)s%(colorreset)s] %(grey)s<%(remote_addr)s - %(url)s - %(endpoint)s>%(colorreset)s - %(module)s::%(funcName)s - %(message)s"


class ArgusRequestLogFormatter(logging.Formatter):
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    blue = "\x1b[34;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    grey = "\x1b[38;2;200;200;200m"
    color_map = {
        logging.DEBUG: grey,
        logging.INFO: blue,
        logging.WARNING: yellow,
        logging.ERROR: red,
        logging.CRITICAL: bold_red
    }

    def format(self, record: logging.LogRecord) -> str:
        record.grey = self.grey
        record.colorreset = self.reset
        record.levelcolor = self.color_map.get(record.levelno, self.grey)
        if has_request_context():
            record.url = request.url
            record.remote_addr = request.remote_addr
            record.endpoint = request.endpoint
        else:
            record.url = ''
            record.remote_addr = ''
            record.endpoint = ''
        return super().format(record)


def setup_application_logging(log_level=logging.INFO):
    dictConfig({
        'version': 1,
        'formatters': {
            'request': {
                'class': f"{__name__}.{ArgusRequestLogFormatter.__name__}",
                'format': LOG_FORMAT_REQUEST,
            }
        },
        'handlers': {
            'main': {
                'class': 'logging.StreamHandler',
                'stream': 'ext://sys.stderr',
                'formatter': 'request'
            }
        },
        'loggers': {
            'cassandra': {
                'level': log_level,
                'handlers': ['main']
            },
            'argus': {
                'level': log_level,
                'handlers': ['main']
            },
            'argus_backend': {
                'level': log_level,
                'handlers': ['main']
            },
            'werkzeug': {
                'level': log_level,
                'handlers': ['main']
            },
            'uwsgi': {
                'level': log_level,
                'handlers': ['main']
            },
            '__main__': {
                'level': log_level,
                'handlers': ['main']
            },
        }
    })



from dataclasses import dataclass


@dataclass(init=True, repr=True)
class Package:
    name: str
    version: str
    date: str
    revision_id: str
    build_id: str


@dataclass(init=True, repr=True)
class LogLink:
    log_name: str
    log_link: str


@dataclass(init=True, repr=True)
class EventsInfo:
    severity: str
    total_events: int
    messages: list[str]



import logging
from datetime import datetime
from uuid import UUID
from time import time

from cassandra.cqlengine import columns
from cassandra.cqlengine.models import Model
from cassandra.cqlengine.usertype import UserType
from flask import Blueprint
from argus.backend.db import ScyllaCluster
from argus.backend.models.web import (
    ArgusTest,
    ArgusGroup,
    ArgusRelease,
    ArgusScheduleGroup,
    ArgusSchedule,
    ArgusScheduleTest,
    ArgusScheduleAssignee,
    User
)
from argus.backend.util.enums import TestInvestigationStatus, TestStatus

LOGGER = logging.getLogger(__name__)


class PluginModelBase(Model):
    _plugin_name = "unknown"
    # Metadata
    build_id = columns.Text(required=True, partition_key=True)
    start_time = columns.DateTime(required=True, primary_key=True, clustering_order="DESC", default=datetime.now)
    id = columns.UUID(index=True, required=True)
    release_id = columns.UUID(index=True)
    group_id = columns.UUID(index=True)
    test_id = columns.UUID(index=True)
    assignee = columns.UUID(index=True)
    status = columns.Text(default=lambda: TestStatus.CREATED.value)
    investigation_status = columns.Text(default=lambda: TestInvestigationStatus.NOT_INVESTIGATED.value)
    heartbeat = columns.Integer(default=lambda: int(time()))
    end_time = columns.DateTime(default=lambda: datetime.utcfromtimestamp(0))
    build_job_url = columns.Text()

    # Test Logs Collection
    logs = columns.List(value_type=columns.Tuple(columns.Text(), columns.Text()))

    @classmethod
    def table_name(cls) -> str:
        return cls.__table_name__

    @classmethod
    def _stats_query(cls) -> str:
        raise NotImplementedError()

    def assign_categories(self):
        key = self.build_id
        try:
            test: ArgusTest = ArgusTest.get(build_system_id=key)
            self.release_id = test.release_id
            self.group_id = test.group_id
            self.test_id = test.id
            if not test.plugin_name:
                test.plugin_name = self._plugin_name
                test.save()
        except ArgusTest.DoesNotExist:
            LOGGER.warning("Test entity missing for key \"%s\", run won't be visible until this is corrected", key)

    def get_scheduled_assignee(self) -> UUID:
        """
            Iterate over all schedules (groups and tests) and return first available assignee
        """
        associated_test = ArgusTest.get(build_system_id=self.build_id)
        associated_group = ArgusGroup.get(id=associated_test.group_id)
        associated_release = ArgusRelease.get(id=associated_test.release_id)

        scheduled_groups = ArgusScheduleGroup.filter(
            release_id=associated_release.id,
            group_id=associated_group.id,
        ).all()

        scheduled_tests = ArgusScheduleTest.filter(
            release_id=associated_release.id,
            test_id=associated_test.id
        ).all()

        unique_schedule_ids = {scheduled_obj.schedule_id for scheduled_obj in [
            *scheduled_tests, *scheduled_groups]}

        schedules = ArgusSchedule.filter(
            release_id=associated_release.id,
            id__in=tuple(unique_schedule_ids),
        ).all()

        today = datetime.utcnow()

        valid_schedules = []
        for schedule in schedules:
            if schedule.period_start <= today <= schedule.period_end:
                valid_schedules.append(schedule)

        assignees_uuids = []
        for schedule in valid_schedules:
            assignees = ArgusScheduleAssignee.filter(
                schedule_id=schedule.id
            ).all()
            assignees_uuids.append(*[assignee.assignee for assignee in assignees])

        return assignees_uuids[0] if len(assignees_uuids) > 0 else None

    @classmethod
    def get_jobs_assigned_to_user(cls, user: User):
        cluster = ScyllaCluster.get()
        query = cluster.prepare("SELECT build_id, start_time, release_id, group_id, assignee, "
                                f"test_id, id, status, investigation_status, build_job_url FROM {cls.table_name()} WHERE assignee = ?")
        rows = cluster.session.execute(query=query, parameters=(user.id,))

        return list(rows)

    @classmethod
    def get_stats_for_release(cls, release: ArgusRelease):
        cluster = ScyllaCluster.get()
        query = cluster.prepare(cls._stats_query())
        rows = cluster.session.execute(query=query, parameters=(release.id,))

        return list(rows)

    @classmethod
    def load_test_run(cls, run_id: UUID) -> 'PluginModelBase':
        raise NotImplementedError()

    @classmethod
    def submit_run(cls, request_data: dict) -> 'PluginModelBase':
        raise NotImplementedError()

    @classmethod
    def get_distinct_product_versions(cls, release: ArgusRelease) -> list[str]:
        raise NotImplementedError()

    def update_heartbeat(self):
        self.heartbeat = int(time())

    def change_status(self, new_status: TestStatus):
        self.status = new_status.value

    def change_investigation_status(self, new_investigation_status: TestInvestigationStatus):
        self.investigation_status = new_investigation_status.value

    def submit_product_version(self, version: str):
        raise NotImplementedError()

    def submit_logs(self, logs: list[dict]):
        raise NotImplementedError()

    def finish_run(self):
        raise NotImplementedError()


class PluginInfoBase:
    # pylint: disable=too-few-public-methods
    name: str
    controller: Blueprint
    model: PluginModelBase
    all_models: list[Model]
    all_types: list[UserType]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['argus',
 'argus.backend',
 'argus.backend.controller',
 'argus.backend.events',
 'argus.backend.models',
 'argus.backend.plugins',
 'argus.backend.plugins.driver_matrix_tests',
 'argus.backend.plugins.sct',
 'argus.backend.service',
 'argus.backend.util',
 'argus.client',
 'argus.client.driver_matrix_tests',
 'argus.client.sct',
 'argus.db']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'scylla-driver>=3.24.8,<4.0.0']

setup_kwargs = {
    'name': 'argus-alm',
    'version': '0.10.1b3',
    'description': 'Argus',
    'long_description': '# Argus\n\n## Description\n\nArgus is a test tracking system intended to provide observability into automated test pipelines which use long-running resources. It allows observation of a test status, its events and its allocated resources. It also allows easy comparison between particular runs of a specific test.\n\n## Installation notes\n\n### Prerequisites\n\n- Python >=3.10.0 (system-wide or pyenv)\n\n- NodeJS >=16 (with npm)\n\n- Yarn (can be installed globally with `npm -g install yarn`)\n\n- nginx\n\n- poetry >=1.2.0b1\n\n### From source\n\n#### Production\n\nPerform the following steps:\n\nCreate a user that will be used by uwsgi:\n\n```bash\nuseradd -m -s /bin/bash argus\nsudo -iu argus\n```\n\n(Optional) Install pyenv and create a virtualenv for this user:\n\n```bash\npyenv install 3.10.0\npyenv virtualenv argus\npyenv activate argus\n```\n\nClone the project into a directory somewhere where user has full write permissions\n\n```bash\ngit clone https://github.com/scylladb/argus ~/app\ncd ~/app\n```\n\nInstall project dependencies:\n\n```bash\npoetry install --with default,dev,web-backend,docker-image\nyarn install\n```\n\nCompile frontend files from `/frontend` into `/public/dist`\n\n```bash\nyarn webpack\n```\n\nCreate a `argus.local.yaml` configuration file (used to configure database connection) and a `argus_web.yaml` (used for webapp secrets) in your application install directory.\n\n```bash\ncp argus_web.example.yaml argus_web.yaml\ncp argus.yaml argus.local.yaml\n```\n\nOpen `argus.local.yaml` and add the database connection information (contact_points, user, password and keyspace name).\n\nOpen `argus_web.yaml` and change the `SECRET_KEY` value to something secure, like a sha512 digest of random bytes. Fill out GITHUB_* variables with their respective values.\n\nCopy nginx configuration file from `docs/configs/argus.nginx.conf` to nginx virtual hosts directory:\n\nUbuntu:\n\n```bash\nsudo cp docs/configs/argus.nginx.conf /etc/nginx/sites-available/argus\nsudo ln -s /etc/nginx/sites-enabled/argus /etc/nginx/sites-available/argus\n```\n\nRHEL/Centos/Alma/Fedora:\n\n```bash\nsudo cp docs/configs/argus.nginx.conf /etc/nginx/conf.d/argus.conf\n```\n\nAdjust the webhost settings in that file as necessary, particularly `listen` and `server_name` directives.\n\nCopy systemd service file from `docs/config/argus.service` to `/etc/systemd/system` directory:\n\n```bash\nsudo cp docs/config/argus.service /etc/systemd/system\n```\n\nOpen it and adjust the path to the `start_argus.sh` script in the `ExecStart=` directive and the user/group, then reload systemd daemon configuration and enable (and optionally start) the service.\n\nWARNING: `start_argus.sh` assumes pyenv is installed into `~/.pyenv`\n\n```bash\nsudo systemctl daemon-reload\nsudo systemctl enable --now argus.service\n```\n\n#### Development\n\nClone the project into a directory somewhere\n\n```bash\ngit clone https://github.com/scylladb/argus\ncd argus\n```\n\nInstall project dependencies:\n\n```bash\npoetry install --with default,dev,web-backend,docker-image\nyarn install\n```\n\nCompile frontend files from `/frontend` into `/public/dist`. Add --watch to recompile files on change.\n\n```bash\nyarn webpack --watch\n```\n##### Configuration\nCreate a `argus.local.yaml` configuration file (used to configure database connection) and a `argus_web.yaml` (used for webapp secrets) in your application install directory.\n\nSee `Production` section for more details.\nTo configure Github authentication follow steps:\n1. Authorize OAuth App\n   1. go to your Account Settings (top right corner) -> Developer settings (left pane) -> OAuth Apps\n   2. Click Create New OAuth App button\n   3. Fill the fields (app name: `argus-dev`, homepage URL `http://localhost:5000`, Auth callback URL: `http://localhost:5000/profile/oauth/github`)\n   4. Confirm and get the tokens/ids required for config\n2. Create Jenkins token for your account\n   1. Go to `Configure` in top right corner\n   2. Click `Add new Token`\n   3. Get it and paste to config to `JENKINS_API_TOKEN` param\n##### Database Initialization\n\nYou can initialize a scylla cluster in any way you like, either using docker image with docker-compose or using cassandra cluster manager. You will need to create the keyspace manually before you can sync database models.\n\nCreate keyspace according to your configuration.\ne.g. (need to test if it works with RF=1 if not, make it 3)\n```\nCREATE KEYSPACE argus WITH replication = {\'class\': \'SimpleStrategy\', \'replication_factor\' : 1}\n```\n\nInitial sync can be done as follows:\n\n```py\nfrom argus.backend.db import ScyllaCluster\nfrom argus.db.testrun import TestRun\ndb = ScyllaCluster.get()\n\ndb.sync_models() # Syncronizes Object Mapper models\nTestRun.init_own_table() # Syncronizes TestRun table (separate from python-driver Object Mapper)\n\n```\n\nYou can also use `flask sync-models` afterwards during development when making small changes to models.\n\nIt is recommended to set up jenkins api key and run `flask scan-jenkins` afterwards to get basic release/group/test structure.\n\nThere are scripts in `./scripts` directory that can be used to download data from production, upload them into your dev db and fix their relations to other models in your instance of the application. Specifically, `download_runs_from_prod.py` requires additional config, `argus.local.prod.yaml` which is the config used to connect to the production cluster. The scripts are split to prevent mistakes and accidentally affecting production cluster.\n\n##### Configuration\n\nCreate a `argus.local.yaml` configuration file (used to configure database connection) and a `argus_web.yaml` (used for webapp secrets) in your application install directory.\n\n```bash\ncp argus_web.example.yaml argus_web.yaml\ncp argus.yaml argus.local.yaml\n```\n\nOpen `argus.local.yaml` and add the database connection information (contact_points, user, password and keyspace name).\n\nOpen `argus_web.yaml` and change the `SECRET_KEY` value to something secure, like a sha512 digest of random bytes. Fill out GITHUB_* and JENKINS_* variables with their respective values.\n\nRun the application from CLI using:\n\n```bash\nFLASK_ENV="development" FLASK_APP="argus_backend:start_server" FLASK_DEBUG=1 CQLENG_ALLOW_SCHEMA_MANAGEMENT=1 flask run\n```\n\nOmit `FLASK_DEBUG` if running your own debugger (pdb, pycharm, vscode)\n',
    'author': 'Alexey Kartashov',
    'author_email': 'alexey.kartashov@scylladb.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/scylladb/argus',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



# TODO: Deprecated, will be removed once REST API client is ready
import logging
import datetime
import time
import traceback
import sys
import threading
from dataclasses import asdict, is_dataclass, fields, Field, dataclass
from typing import Any
from uuid import uuid4, UUID

from argus.db.config import BaseConfig
from argus.db.utils import is_list_homogeneous
from argus.db.cloud_types import CloudResource, CloudInstanceDetails, BaseCloudSetupDetails
from argus.db.interface import ArgusDatabase
from argus.db.db_types import ColumnInfo, CollectionHint, NemesisRunInfo, TestStatus, TestInvestigationStatus, \
    EventsBySeverity, PackageVersion
from argus.backend.models.web import ArgusRelease, ArgusGroup, ArgusTest, ArgusSchedule, ArgusScheduleAssignee, ArgusScheduleGroup, \
    ArgusScheduleTest

LOGGER = logging.getLogger(__name__)


class TestInfoSerializationError(Exception):
    pass


class TestInfoSchemaError(Exception):
    pass


class TestInfoValueError(Exception):
    pass


class BaseTestInfo:
    EXPOSED_ATTRIBUTES = {}
    ATTRIBUTE_CONSTRAINTS = {}
    COLLECTION_HINTS = {}

    def __init__(self, *args, **kwargs):
        pass

    @classmethod
    def create_skeleton(cls):
        pass

    @classmethod
    def schema(cls):
        data = {}
        for attr, column_type in cls.EXPOSED_ATTRIBUTES.items():
            value = None
            if column_type is list or column_type is tuple:
                value = cls.schema_process_collection(attr)
                column_type = CollectionHint
            constraints = cls.ATTRIBUTE_CONSTRAINTS.get(attr, [])
            column_info = ColumnInfo(
                name=attr, type=column_type, value=value, constraints=constraints)
            data[attr] = column_info

        return data

    @classmethod
    def schema_process_collection(cls, attr_name: str):
        hint = cls.COLLECTION_HINTS.get(attr_name)
        if not hint:
            raise TestInfoSchemaError(
                "Encountered a collection and no collection hint was found")

        return hint

    def serialize(self):
        data = {}
        for attr in self.EXPOSED_ATTRIBUTES:
            attribute_value = getattr(self, attr)
            if isinstance(attribute_value, list):
                attribute_value = self._process_list(attribute_value)
            elif is_dataclass(attribute_value):
                attribute_value = asdict(attribute_value)

            data[attr] = attribute_value

        return data

    @staticmethod
    def _process_list(list_to_check: list[Any]):
        if len(list_to_check) == 0:
            return list_to_check

        if not is_list_homogeneous(list_to_check):
            raise TestInfoSerializationError("Detected a non-homogenous list")

        contains_dataclass = is_dataclass(list_to_check[0])

        if contains_dataclass:
            return [asdict(dc) for dc in list_to_check]

        return list_to_check

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value, field):  # pylint: disable=unused-argument
        return value


class TestDetails(BaseTestInfo):
    # pylint: disable=too-many-instance-attributes
    EXPOSED_ATTRIBUTES = {"scm_revision_id": str, "started_by": str, "build_job_url": str,
                          "start_time": datetime.datetime, "end_time": datetime.datetime,
                          "config_files": list, "packages": list, "scylla_version": str,
                          "yaml_test_duration": int,
                          }
    COLLECTION_HINTS = {
        "packages": CollectionHint(list[PackageVersion]),
        "config_files": CollectionHint(list[str]),
    }

    def __init__(self, scm_revision_id: str,
                 started_by: str, build_job_url: str,
                 yaml_test_duration: int, start_time: datetime,
                 config_files: list[str], packages: list[PackageVersion],
                 end_time: datetime.datetime = datetime.datetime.utcfromtimestamp(0),
                 scylla_version: str | None = None):
        # pylint: disable=too-many-arguments
        super().__init__()
        self.scm_revision_id = scm_revision_id
        self.started_by = started_by
        self.build_job_url = build_job_url
        self.start_time = start_time
        self.yaml_test_duration = yaml_test_duration
        if not (is_list_homogeneous(packages) or (
                len(packages) > 0 and isinstance(next(iter(packages)), PackageVersion))):
            raise TestInfoValueError(
                "Package list contains incorrect values", packages)
        self.packages = packages
        self.config_files = config_files
        self.end_time = end_time
        self.scylla_version = scylla_version

    @classmethod
    def from_db_row(cls, row):
        if row.packages:
            packages = [PackageVersion.from_db_udt(
                udt) for udt in row.packages]
        else:
            packages = []

        config_files = row.config_files if row.config_files else []

        return cls(scm_revision_id=row.scm_revision_id, started_by=row.started_by, build_job_url=row.build_job_url,
                   start_time=row.start_time, end_time=row.end_time, yaml_test_duration=row.yaml_test_duration,
                   config_files=config_files, scylla_version=row.scylla_version,
                   packages=packages)

    def set_test_end_time(self):
        self.end_time = datetime.datetime.utcnow().replace(microsecond=0)


class TestResourcesSetup(BaseTestInfo):
    EXPOSED_ATTRIBUTES = {
        "sct_runner_host": CloudInstanceDetails,
        "region_name": list,
        "cloud_setup": BaseCloudSetupDetails
    }
    COLLECTION_HINTS = {
        "region_name": CollectionHint(list[str]),
    }

    def __init__(self, sct_runner_host: CloudInstanceDetails,
                 region_name: list[str], cloud_setup: BaseCloudSetupDetails):
        super().__init__()
        self.sct_runner_host = sct_runner_host
        self.region_name = region_name
        self.cloud_setup = cloud_setup

    @classmethod
    def from_db_row(cls, row):
        runner = CloudInstanceDetails.from_db_udt(row.sct_runner_host)
        cloud_setup = BaseCloudSetupDetails.from_db_udt(row.cloud_setup)

        regions = row.region_name if row.region_name else []

        return cls(sct_runner_host=runner, region_name=regions,
                   cloud_setup=cloud_setup)


class TestLogs(BaseTestInfo):
    EXPOSED_ATTRIBUTES = {"logs": list}
    COLLECTION_HINTS = {
        "logs": CollectionHint(list[tuple[str, str]])
    }

    def __init__(self):
        super().__init__()
        self._log_collection = []

    def add_log(self, log_type: str, log_url: str | list[str]) -> None:
        if isinstance(log_url, str):
            self._log_collection.append((log_type, log_url))
        elif isinstance(log_url, list):
            for log in log_url:
                self._log_collection.append((log_type, log))
        else:
            LOGGER.warning("Unknown log type encountered: %s", log_url)

    @property
    def logs(self) -> list[tuple[str, str]]:
        return self._log_collection

    @classmethod
    def from_db_row(cls, row):
        logs = cls()
        if row.logs:
            for log_type, log_url in row.logs:
                logs.add_log(log_type, log_url)

        return logs


class TestResources(BaseTestInfo):
    EXPOSED_ATTRIBUTES = {"allocated_resources": list}
    COLLECTION_HINTS = {
        "allocated_resources": CollectionHint(list[CloudResource]),
    }

    def __init__(self):
        super().__init__()
        self._allocated_resources = []

    def attach_resource(self, resource: CloudResource):
        self._allocated_resources.append(resource)
        self._allocated_resources.sort(key=lambda v: v.name)

    def detach_resource(self, resource: CloudResource, reason: str = "unspecified reason"):
        resource_to_detach = next(
            r for r in self._allocated_resources if r == resource)
        resource_to_detach.terminate(reason=reason)

    @property
    def allocated_resources(self) -> list[CloudResource]:
        return self._allocated_resources

    @classmethod
    def from_db_row(cls, row):
        resources = cls()
        resource_row = row.allocated_resources if row.allocated_resources else []
        for resource in resource_row:
            cloud_resource = CloudResource.from_db_udt(resource)
            resources.allocated_resources.append(cloud_resource)
        resources.allocated_resources.sort(key=lambda v: v.name)
        return resources


class TestResults(BaseTestInfo):
    # pylint: disable=too-many-arguments
    EXPOSED_ATTRIBUTES = {"status": str, "events": list,
                          "nemesis_data": list, "screenshots": list}
    COLLECTION_HINTS = {
        "events": CollectionHint(list[EventsBySeverity]),
        "nemesis_data": CollectionHint(list[NemesisRunInfo]),
        "screenshots": CollectionHint(list[str]),
    }

    def __init__(self, status: TestStatus, events: list[EventsBySeverity] = None,
                 nemesis_data: list[NemesisRunInfo] = None, screenshots: list[str] = None,
                 max_stored_events=25):
        super().__init__()
        if isinstance(status, TestStatus):
            self._status = status.value
        else:
            self._status = TestStatus(status).value
        self.events = events if events else []
        self.nemesis_data = nemesis_data if nemesis_data else []
        self.screenshots = screenshots if screenshots else []
        self.max_stored_events = max_stored_events

    @classmethod
    def from_db_row(cls, row):
        if row.events:
            events = [EventsBySeverity.from_db_udt(
                event) for event in row.events]
        else:
            events = []

        if row.nemesis_data:
            nemesis_data = [NemesisRunInfo.from_db_udt(
                nemesis) for nemesis in row.nemesis_data]
        else:
            nemesis_data = []

        if row.screenshots:
            screenshots = row.screenshots
        else:
            screenshots = []

        return cls(status=row.status, events=events, nemesis_data=nemesis_data, screenshots=screenshots)

    def _add_new_event_type(self, event: EventsBySeverity):
        if len([v for v in self.events if v.severity == event.severity]) > 0:
            raise TestInfoValueError(
                f"Severity event collection {event.severity} already exists in TestResults")

        self.events.append(event)

    def _collect_event_message(self, event: EventsBySeverity, message: str):
        if len(event.last_events) >= self.max_stored_events:
            event.last_events = event.last_events[1:]

        event.event_amount += 1
        event.last_events.append(message)

    def add_nemesis(self, nemesis: NemesisRunInfo):
        self.nemesis_data.append(nemesis)

    def add_event(self, event_severity: str, event_message: str):
        try:
            event = next(filter(lambda v: v.severity ==
                         event_severity, self.events))
        except StopIteration:
            event = EventsBySeverity(
                severity=event_severity, event_amount=0, last_events=[])
            self._add_new_event_type(event)

        self._collect_event_message(event, event_message)

    def add_screenshot(self, screenshot_link: str):
        self.screenshots.append(screenshot_link)

    @property
    def status(self) -> TestStatus:
        return TestStatus(self._status)

    @status.setter
    def status(self, value: TestStatus):
        self._status = TestStatus(value).value


@dataclass
class TestRunInfo:
    details: TestDetails
    setup: TestResourcesSetup
    resources: TestResources
    logs: TestLogs
    results: TestResults


class TestRun:
    # pylint: disable=too-many-instance-attributes
    EXPOSED_ATTRIBUTES = {"id": UUID, "group_id": UUID, "release_id": UUID,
                          "build_id": str, "test_id": UUID,
                          "assignee": UUID, "heartbeat": int, "investigation_status": str}
    ATTRIBUTE_CONSTRAINTS = {
    }
    PRIMARY_KEYS = {
        "build_id": (str, "partition"),
        "start_time": (datetime.datetime, "clustering"),
    }
    CLUSTERING_ORDER = {
        "start_time": "DESC",
    }
    INDICES = ["release_id", "group_id", "test_id", "id", "assignee", "status"]
    _USING_RUNINFO = TestRunInfo
    _TABLE_NAME = "test_runs_v8"
    _IS_TABLE_INITIALIZED = False
    _ARGUS_DB_INTERFACE = None

    def __init__(self, test_id: UUID, assignee: UUID, build_id: str,
                 run_info: TestRunInfo, config: BaseConfig = None, argus_interface: ArgusDatabase = None,
                 investigation_status: str = TestInvestigationStatus.NOT_INVESTIGATED):
        # pylint: disable=too-many-arguments
        if not test_id:
            test_id = uuid4()
        self._save_lock = threading.Lock()
        self._id = test_id
        self._build_id = build_id
        self._group_id = None
        self._release_id = None
        self._test_id = None
        self._assignee = assignee
        self._investigation_status = investigation_status
        self._run_info = run_info
        self._heartbeat = int(time.time())
        self._config = config
        for field in fields(run_info):
            setattr(self, field.name, getattr(run_info, field.name))

        if argus_interface:
            self.argus = argus_interface

    @classmethod
    def table_name(cls) -> str:
        return cls._TABLE_NAME

    @classmethod
    def from_db_row(cls, row, config: BaseConfig = None):
        if not cls._IS_TABLE_INITIALIZED:
            cls.init_own_table(config)
        nested_fields = {}
        for field in fields(cls._USING_RUNINFO):
            nested_fields[field.name] = field.type.from_db_row(row)

        run_info = cls._USING_RUNINFO(**nested_fields)
        run = cls(test_id=row.id, assignee=row.assignee, build_id=row.build_id,
                  run_info=run_info, investigation_status=row.investigation_status)
        run.heartbeat = row.heartbeat
        run.group_id = row.group_id
        run.release_id = row.release_id
        run.test_id = row.test_id
        return run

    @classmethod
    def from_id(cls, test_id: UUID, config: BaseConfig = None):
        if not cls._IS_TABLE_INITIALIZED:
            cls.init_own_table(config)
        database = cls.get_argus()
        if row := database.fetch(cls._TABLE_NAME, test_id):
            return cls.from_db_row(row)

        return None

    @classmethod
    def from_pk(cls, pk: tuple[str, datetime.datetime], config: BaseConfig = None):
        if not cls._IS_TABLE_INITIALIZED:
            cls.init_own_table(config)
        database = cls.get_argus()
        if row := database.fetch_generic(cls._TABLE_NAME, pk, "WHERE build_id = ? and start_time = ?"):
            return cls.from_db_row(row)

        return None

    @classmethod
    def get_argus(cls, config: BaseConfig = None) -> ArgusDatabase:
        if not cls._ARGUS_DB_INTERFACE:
            cls._ARGUS_DB_INTERFACE = ArgusDatabase(config=config)
        return cls._ARGUS_DB_INTERFACE

    @classmethod
    def set_argus(cls, argus_interface: ArgusDatabase):
        cls._ARGUS_DB_INTERFACE = argus_interface
        cls._IS_TABLE_INITIALIZED = False

    @property
    def argus(self) -> ArgusDatabase:
        if not self._ARGUS_DB_INTERFACE:
            self.get_argus(self._config)
        return self._ARGUS_DB_INTERFACE

    @argus.setter
    def argus(self, interface: ArgusDatabase | None):
        self._ARGUS_DB_INTERFACE = interface  # pylint: disable=invalid-name
        self._IS_TABLE_INITIALIZED = False  # pylint: disable=invalid-name

    @property
    def heartbeat(self) -> int:
        return self._heartbeat

    @heartbeat.setter
    def heartbeat(self, value: int | float):
        self._heartbeat = int(value)

    @property
    def build_id(self) -> str:
        return self._build_id

    @build_id.setter
    def build_id(self, value: str) -> None:
        self._build_id = str(value)

    @property
    def id(self) -> UUID:  # pylint: disable=invalid-name
        return self._id

    @property
    def group_id(self) -> UUID:
        return self._group_id

    @property
    def release_id(self) -> UUID:
        return self._release_id

    @property
    def test_id(self) -> UUID:
        return self._test_id

    @release_id.setter
    def release_id(self, value: UUID) -> None:
        self._release_id = value

    @group_id.setter
    def group_id(self, value: UUID) -> None:
        self._group_id = value

    @test_id.setter
    def test_id(self, value: UUID) -> None:
        self._test_id = value

    @property
    def assignee(self) -> UUID:
        return self._assignee

    @assignee.setter
    def assignee(self, value):
        self._assignee = value

    @property
    def investigation_status(self) -> str:
        return self._investigation_status

    @investigation_status.setter
    def investigation_status(self, value: TestInvestigationStatus | str):
        self._investigation_status = TestInvestigationStatus(value)

    def serialize(self) -> dict[str, Any]:
        LOGGER.debug("Serializing test run...")
        nested_data = {}
        for field in fields(self._USING_RUNINFO):
            field: Field
            value: BaseTestInfo = getattr(self, field.name)
            nested_data = {
                **nested_data,
                **value.serialize()
            }

        data = {
            "build_id": self._build_id,
            "id": self._id,
            "group_id": self._group_id,
            "release_id": self._release_id,
            "test_id": self._test_id,
            "assignee": self._assignee,
            "heartbeat": self._heartbeat,
            "investigation_status": self._investigation_status,
            **nested_data
        }
        LOGGER.debug("Serialized Data: %s", data)
        return data

    @classmethod
    def init_own_table(cls, config: BaseConfig = None):
        LOGGER.debug("Initializing TestRun table...")
        cls.get_argus(config).init_table(
            table_name=cls._TABLE_NAME, column_info=cls.schema())
        cls._IS_TABLE_INITIALIZED = True

    @classmethod
    def set_table_name(cls, new_table_name: str):
        cls._TABLE_NAME = new_table_name
        cls._IS_TABLE_INITIALIZED = False

    @classmethod
    def schema(cls) -> dict[str, ColumnInfo]:
        data = {}
        LOGGER.debug("Dumping full schema...")
        for attr, column_type in cls.EXPOSED_ATTRIBUTES.items():
            value = None
            constraints = cls.ATTRIBUTE_CONSTRAINTS.get(attr, [])
            column_info = ColumnInfo(
                name=attr, type=column_type, value=value, constraints=constraints)
            data[attr] = column_info

        schema_dump = {}
        for field in fields(cls._USING_RUNINFO):
            schema_dump = {
                **schema_dump,
                **field.type.schema(),
            }

        full_schema = dict(
            **{"$tablekeys$": cls.PRIMARY_KEYS},
            **{"$clustering_order$": cls.CLUSTERING_ORDER},
            **{"$indices$": cls.INDICES},
            **data,
            **schema_dump
        )
        LOGGER.debug("Full Schema: %s", full_schema)
        return full_schema

    def save(self):
        with self._save_lock:
            if not self._IS_TABLE_INITIALIZED:
                self.init_own_table(self._config)
            if not self.exists():
                self._assign_categories()
                if not self.assignee:
                    try:
                        self.assignee = self._get_current_assignee_from_schedule()
                    except Exception:  # pylint: disable=broad-except
                        LOGGER.warning("Error getting assignee from database")
                        LOGGER.debug("Details: ", exc_info=True)

                LOGGER.debug("Inserting data for test run: %s", self.id)
                self.argus.insert(table_name=self._TABLE_NAME,
                                  run_data=self.serialize())
            else:
                LOGGER.debug("Updating data for test run: %s", self.id)
                self.argus.update(table_name=self._TABLE_NAME,
                                  run_data=self.serialize())

    def exists(self) -> bool:
        if not self._IS_TABLE_INITIALIZED:
            self.init_own_table(self._config)

        if self.argus.fetch(table_name=self._TABLE_NAME, run_id=self.id):
            return True
        return False

    def _assign_categories(self):
        key = self._build_id
        try:
            test: ArgusTest = ArgusTest.using(
                connection=self.argus.CQL_ENGINE_CONNECTION_NAME
            ).get(build_system_id=key)
            self.release_id = test.release_id
            self.group_id = test.group_id
            self.test_id = test.id
        except ArgusTest.DoesNotExist:
            LOGGER.warning(
                "Test entity missing for key \"%s\", run won't be visible until this is corrected", key)

    def _get_current_assignee_from_schedule(self) -> UUID:
        """
            Iterate over all schedules (groups and tests) and return first available assignee
        """
        associated_test = ArgusTest.using(
            connection=self.argus.CQL_ENGINE_CONNECTION_NAME
        ).get(build_system_id=self.build_id)
        associated_group = ArgusGroup.using(
            connection=self.argus.CQL_ENGINE_CONNECTION_NAME
        ).get(id=associated_test.group_id)
        associated_release = ArgusRelease.using(
            connection=self.argus.CQL_ENGINE_CONNECTION_NAME
        ).get(id=associated_test.release_id)

        scheduled_groups = ArgusScheduleGroup.filter(
            release_id=associated_release.id, group_id=associated_group.id
        ).all().using(
            connection=self.argus.CQL_ENGINE_CONNECTION_NAME
        )

        scheduled_tests = ArgusScheduleTest.filter(
            release_id=associated_release.id, test_id=associated_test.id
        ).all().using(
            connection=self.argus.CQL_ENGINE_CONNECTION_NAME
        )

        unique_schedule_ids = {scheduled_obj.schedule_id for scheduled_obj in [
            *scheduled_tests, *scheduled_groups]}

        schedules = ArgusSchedule.filter(
            release_id=associated_release.id, id__in=tuple(
                unique_schedule_ids)
        ).all().using(
            connection=self.argus.CQL_ENGINE_CONNECTION_NAME
        )

        today = datetime.datetime.utcnow()

        valid_schedules = []
        for schedule in schedules:
            if schedule.period_start <= today <= schedule.period_end:
                valid_schedules.append(schedule)

        assignees_uuids = []
        for schedule in valid_schedules:
            assignees = ArgusScheduleAssignee.filter(
                schedule_id=schedule.id
            ).all().using(
                connection=self.argus.CQL_ENGINE_CONNECTION_NAME
            )
            assignees_uuids.append(
                *[assignee.assignee for assignee in assignees])

        return assignees_uuids[0] if len(assignees_uuids) > 0 else None

    def shutdown(self):
        LOGGER.debug("Shutting down cluster connection...")
        self.argus.cluster.shutdown()

    @property
    def run_info(self) -> TestRunInfo:
        return self._run_info


class TestRunWithHeartbeat(TestRun):
    def __init__(self, test_id: UUID, assignee: UUID, build_id: str,
                 run_info: TestRunInfo, heartbeat_interval=30, config: BaseConfig = None,
                 argus_interface: ArgusDatabase = None, investigation_status: str = TestInvestigationStatus.NOT_INVESTIGATED,):
        # pylint: disable=too-many-arguments
        self._heartbeat_interval = heartbeat_interval
        self._shutdown_event = threading.Event()
        super().__init__(test_id=test_id, assignee=assignee, build_id=build_id,
                         investigation_status=investigation_status, run_info=run_info,
                         config=config, argus_interface=argus_interface)
        self._thread = threading.Thread(target=self._heartbeat_entry,
                                        name=f"{self.__class__.__name__}-{self.id}-heartbeat", daemon=True)
        self._heartbeat_statement = self.argus.session.prepare(
            f"UPDATE {TestRun.table_name()} SET heartbeat = ? WHERE build_id = ? AND start_time = ?")
        self._thread.start()

    @property
    def heartbeat_interval(self) -> int:
        return self._heartbeat_interval

    @heartbeat_interval.setter
    def heartbeat_interval(self, value: float | int):
        self._heartbeat_interval = value

    @property
    def thread(self):
        return self._thread

    def _heartbeat_entry(self):
        while True:
            time.sleep(self.heartbeat_interval)
            if self._shutdown_event.is_set():
                break
            LOGGER.debug("Sending heartbeat...")
            self.heartbeat = time.time()
            bound_statement = self._heartbeat_statement.bind(
                (self.heartbeat, self.build_id, self.run_info.details.start_time))
            self.argus.session.execute(bound_statement)
        LOGGER.debug("Heartbeat exit")

    def shutdown(self):
        self._shutdown_event.set()
        LOGGER.debug("Waiting for the heartbeat thread to exit...")
        self._thread.join(timeout=self.heartbeat_interval + 10)
        if self._thread.is_alive():
            LOGGER.warning(
                "Heartbeat thread was not able to shut down correctly. Stack trace:")
            # pylint: disable=protected-access
            current_threads = sys._current_frames()
            stack_trace = traceback.extract_stack(
                current_threads[self._thread.ident])
            LOGGER.warning(
                "\n".join([f'#{lineno:3} : {line:50}: {fname}' for fname, lineno, _, line in stack_trace]))
        super().shutdown()



from datetime import datetime
import json
from uuid import UUID


class ArgusJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        match obj:
            case UUID():
                return str(obj)
            case datetime():
                return obj.isoformat(sep=' ', timespec='milliseconds')
            case _:
                return json.JSONEncoder.default(self, obj)
